##############################################################################
#
# RROC-function.r: part of the RROCcurves library
#
# RROCcurves library
#  JOSE HERNANDEZ-ORALLO, UNIVERSITAT POLITECNICA DE VALENCIA, SPAIN
#  jorallo@dsic.upv.es
#
# See "INFO.TXT" for copyright information, history, related papers, etc.
#
##############################################################################

##############################################################################
################ FUNCTIONS PROVIDED AND A SHORT EXPLANATION ##################
############## (more details inside each function definition) ################
##############################################################################
#
# RROC_space(...): Plots the underlying RROC space (use this just before plotting any model)
#
# RROC_curve(...) : Returns a RROC curve of a regression model and (optionally) plots it on a previously created RROC space
#
# RROC_convexhull(...): Returns convex hull of a set of points in RROC space and (optionally) plots it on a previously created RROC space
#
# RROC_intercept(...): Intercepts a cost context (alpha) on a RROC curve (assumes the curve is convex)
#
# RROC_intercept_many(...): Intercepts a cost context (alpha) on several RROC curve (as RROC_intercept_many) but for several curves in a list
#
# RROC_dominance_intervals(...): Finds the models that dominate (and their ranges) for several curves in a list
#
# REC_space(...): Plots the underlying REC space (use this just before plotting any model)
#
# REC_curve(...) :  Returns a REC curve of a regression model and (optionally) plots it on a previously created REC space
#
# loss_linlin(...): Returns the asymmetric absolute error loss (linlin)
#
# calibrate_regressor(...): Calibrates a regression model by adding a constant shift to get 0 bias (overpredictions = underpredictions). It has no effect on RROC curves but it does on REC curves
#
# optimal_shift_linlin (...): Returns the best shift for a regression model (wrt. the linlin loss determined by an alpha)
#
##############################################################################




######################
##### FUNCTIONS ######
######################




#### RROC_space ####
#
# Plots the RROC space
#
#### Arguments:
# normalise_axes: divide x-axis and y-axis by N
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
# xlim0, ylim0: part of the space shown
# DRAW_DIAGONAL: if true draws a diagonal on the RROC space
#
RROC_space <- function(normalise_axes=FALSE, N, xlim0, ylim0, DRAW_DIAGONAL) {

  if (!normalise_axes) {
    N <- 1 
    xlab <- "OVER"
    ylab <- "UNDER"
  } else {
    xlab <- "OVER / n"
    ylab <- "UNDER / n"
  }

  xlim0 <- xlim0 / N
  ylim0 <- ylim0 / N

  OVER=0
  UNDER=0

  plot(OVER, UNDER, lty="blank", col="white", xlim=c(0,xlim0), ylim=(c(-ylim0,0)), xlab=xlab, ylab=ylab)  #xlim 

  # draws diagonal
  if (DRAW_DIAGONAL) { 
    lines(c(-xlim0*2,xlim0*2), c(ylim0*2,-ylim0*2), col="black", lty="dashed") # 0=blank, 1=solid, 2=dashed, 3=dotted, 4=dotdash, 5=longdash, 6=twodash
  }
}


#### RROC_curve ####
#
# Returns a RROC curve of a regression model and (optionally) plots it on a previous created RROC space
#
#### Arguments:
# predicted: an array of predicted values
# actual: an array of true values (size of predicted and actual should match)
# mytext: text used to accompany the curve. If empty nothing is shown.
# plot_point: if TRUE, it plots the point which corresponds to shift=0 (the crisp regression model).
# plot_rectangle: if TRUE, shows the rectangle from the point which corresponds to shift=0 to RROC heaven (0,0). The perimeter of this rectangle equals twice the absolute error.
# plot_curve: if TRUE, plots the RROC curve on a previously created RROC curve. Set to FALSE if you only want to get the points of curve.
# colour1: colour used for points, rectangles and old way of calculating the curve (still used for other shift functions)
# colour2: colour used for the curve
# shiftfunction: we can provide a function to calculate the shift. If NULL, the usual RROC curves is plotted.
# normalise_axes: divide x-axis and y-axis by N
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
# xlim: size of the RROC space in the x-axis
# ylim: size of the RROC space in the x-axis
# USE_SD: if TRUE it uses SD to calculate a relative shift. It is FALSE by default.
# DRAW_CIRCLES: if TRUE it shows a small circle for each point. It is TRUE by default.
# PLOT_BIAS: if TRUE it plots a vertical and a horizontal line to the diagonal. Any of both shows the Error Bias
# PLOT_SQBIAS: if TRUE it draws a shaded triangle with the 1/2 the squared error bias
# PLOT_USE: if TRUE it plots a line to RROC heaven, which equals the unweighted Squared Error (uSE)
# PLOT_AE: if TRUE it plots two lines horizontal to the left and vertical to RROC heaven. The sum of these two lines equals the absolute error (AE)
# my_lty : line style for the curve

#
#### Returns
# The curve as a bidimensional matrix with the X,Y points of the curve.
#
RROC_curve <- function(predicted, actual, mytext, plot_point, plot_rectangle, plot_curve, colour1, colour2, shiftfunction=NULL, normalise_axes=FALSE, N, xlim0, ylim0, USE_SD=FALSE, DRAW_CIRCLES= TRUE, PLOT_BIAS=FALSE, PLOT_SQBIAS= FALSE, PLOT_USE= FALSE, PLOT_AE= FALSE, my_lty=1) {



  nex <- length(predicted)                              # Number of examples
  orig_error <- predicted-actual                        # Calculates residuals (error vector)
  
  error_range <- max(orig_error) - min(orig_error)      # Calculates the error range

  print("*************************************")
  print("RROC_curve function: Regression model:")
#  print(rbind(predicted,actual,orig_error))

  orig_over <- sum(orig_error[which(orig_error > 0)])   # Calculates OVER
  orig_under <- sum(orig_error[which(orig_error <= 0)]) # Calculates UNDER

  print("Original OVER and UNDER")
  print(orig_over)
  print(orig_under)

  mae <- mean(abs(orig_error))                          # Calculates Mean Absolute Error
  mse <- mean(orig_error^2)                             # Calculates Mean Squared Error

  print("mae and mse")
  print(mae)
  print(mse)
  print("bias")
  print(mean(orig_error))


  if (!normalise_axes)
    N <- 1

  # In case we want to plot the 0-bias point and rectangle
  # shift=0
  if (plot_point) {
    points(orig_over/N, orig_under/N, pch=7, col=colour1)  # pch=2, 7, 15, circles: 19, 21
  }
  if (plot_rectangle) {
    lines(c(orig_over/N,orig_over/N,xlim0*2), c(-ylim0*2, orig_under/N,orig_under/N), col=colour1) 
  }


  if (PLOT_BIAS) { # if TRUE it plots a vertical and a horizontal line to the diagonal. Any of both shows the Error Bias
    lines(c(orig_over/N,orig_over/N), c(orig_under/N, -orig_over/N), col="orange")  # vertical to the diagonal = Error Bias
    lines(c(orig_over/N,-orig_under/N), c(orig_under/N, orig_under/N), col="orange")  # horizontal to the diagonal = Error Bias
  }
  if (PLOT_SQBIAS) { # if TRUE it draws a shaded triangle with the 1/2 the squared error bias
    polygon(c(orig_over/N,orig_over/N,-orig_under/N,orig_over/N), c(orig_under/N, -orig_over/N, orig_under/N, orig_under/N), density=20, col="orange") # shaded area = 1/2 squared error bias
  }

  if (PLOT_USE) { # if TRUE it plots a line to RROC heaven, which equals the unweighted Squared Error (uSE)
    lines(c(orig_over/N,0), c(orig_under/N, 0), col="violet", lty="dotted", lwd=2)  # to RROC heaven = uSE
  }

  if (PLOT_AE) { # if TRUE it plots two lines horizontal to the left and vertical to RROC heaven. The sum of these two lines equals the absolute error (AE)
    lines(c(orig_over/N,0,0), c(orig_under/N, orig_under/N,0), col="brown", lty="dotdash")  # horizontal to the left and vertical to RROC heaven = AE
  }
   # "blank", "solid", "dashed", "dotted", "dotdash", "longdash", or "twodash"



  # Writes a text (e.g., the name of the model) if mytext is not null
  if (!is.null(mytext)) { 
      text((orig_over + xlim0/20)/N, (orig_under - ylim0/66.666666)/N, mytext, col=colour1, cex=0.75)
#       text((orig_over + xlim0/20)/N, (orig_under - ylim0/40)/N, mytext, col=colour1, cex=0.75) # This one for fig2
  }


  # Here we decide which method we use to draw the curve. 
  if (is.null(shiftfunction)) { # If the shiftfunction is not specified (usual RROC curves) then the SEGMENTWISEMETHOD is used (the algorithm in the paper)
    POINTWISEMETHOD <- FALSE
    POINTWISECURVE <- FALSE # This can only be TRUE if POINTWISEMETHOD is TRUE
    SEGMENTWISEMETHOD <- TRUE
  } else {  # This method just samples the space. It is valid for any shift function, but plots are constructed point by point and calculations are approximate.
    POINTWISEMETHOD <- TRUE
    POINTWISECURVE <- TRUE # This can only be TRUE if POINTWISEMETHOD is TRUE
    SEGMENTWISEMETHOD <- FALSE
  }


  ### METHOD 1 (NOT-COMMONLY USED NOW, ONLY FOR NON-USUAL (NON-OPTIMAL) SHIFT FUNCTIONS) 
  if (POINTWISEMETHOD) {  # This method is used when the shift function is not constant
                          # This was the old method before developing the algorithm and now it is only used for special shift functions.

    RESO <- 2000  # This the resolution used. The greater the finer the curve is (the slower it is to draw, too)
    mins <- -error_range*1.01 # the lowest shift we will use (a little bit higher, 1%, than the error_range)
    maxs <- +error_range*1.01 # the highest shift we will use (a little bit higher, 1%, than the error_range)
    shift <- (((1:RESO) - RESO/2)/ (RESO)) * (maxs-mins)  # how the shifts range 

    under <- NULL  # This vector accumulates underestimations and it is used to draw the curve
    over <- NULL   # This vector accumulates overestimations and it is used to draw the curve
    shmae <- NULL
    shmse <- NULL
    underc <- -1
    overc <- -1

    ROCpointsSW_X <- NULL  # This vector finally gets the curve (X)
    ROCpointsSW_Y <- NULL  # This vector finally gets the curve (X)
    l <- 1
    ROCpointsSW_X[l] <- 0
    ROCpointsSW_Y[l] <- Inf

    for (i in 1:RESO) {

      # Use confidence modifier
      if (USE_SD) {
        prednew <- predicted + shift[i] * sd
      } else {
         if (is.null(shiftfunction)) {
            prednew <- predicted + shift[i]  # Constant shift
         } else {
           prednew <- shiftfunction(predicted, shift[i])
         }   
      }

     error <- prednew - actual
     newl <- FALSE

     overv <- error[which(error > 0)]
     c <- length(overv)
     if (overc != c) {
       newl <- TRUE
     }
     overc <- c
     if (overc == 0) {
       over[i] <- 0
     } else {
       over[i] <- sum(overv)
     }

     underv <- error[which(error <= 0)]   # ! lower than equal here
     c <- length(underv)
     if (underc != c) {
       newl <- TRUE
     }
     underc <- c
     if (underc == 0) {
       under[i] <- 0
     } else {
       under[i] <- sum(underv)
     }

     if (newl) {
       l <- l+1
       ROCpointsSW_X[l] <- over[i]
       ROCpointsSW_Y[l] <- under[i]
     }

     shmae[i] <- (over[i] + under[i]) / nex
     shmse[i] <- (over[i] + under[i])^2 / nex 
    }

    if (!newl) {
      l <- l+1
      ROCpointsSW_X[l] <- over[i]
      ROCpointsSW_Y[l] <- under[i]
    }
    l <- l+1
    ROCpointsSW_X[l] <- -Inf
    ROCpointsSW_Y[l] <- 0


    AreaPW <- 0
    AccLength <- 0
    for (i in 2:RESO) {
     Length <- -(under[i]-under[i-1])
     AccLength <- AccLength + Length
     AreaPW <- AreaPW - (Length)*(over[i] + over[i-1])/2 
    }

    print("Without ties: lines = nex + 2 + 2(infty lines)")
    print(length(ROCpointsSW_X))
    print(nex)
    print("Area (pointwise, i.e., approximate):")
    print(AreaPW)

    if (plot_curve && POINTWISECURVE) {
      lines(over/N, under/N, col=colour1)
    }
  }  # END POINTWISEMETHOD





  # A BETTER WAY TO DRAW THE CURVES FOR THE USUAL (OPTIMAL) SHIFT FUNCTION
  if (SEGMENTWISEMETHOD) {  # This way can only be used when the shift function is constant and chosen optimally
    errors <- sort(orig_error, decreasing=TRUE)  # sorts the errors
    n <- length(errors)
    bignumber <- sum(abs(errors))*1000 # We use a big number according to the scale of the errors (we cannot use infinite to draw)

    # Draws the curve from bottom-left corner to top-right corner
    ROCpointsSW_X <- NULL
    ROCpointsSW_Y <- NULL
    ROCpointsSW_X[1] <- 0
    ROCpointsSW_Y[1] <- -bignumber # -Inf
    for(i in 1:(n)) {
      errorsi <- errors - errors[i]         # errors[i] is the shift
      ROCpointsSW_X[i+1] <- sum(errorsi[which(errorsi > 0)])
      ROCpointsSW_Y[i+1] <- sum(errorsi[which(errorsi <= 0)])
    }
    ROCpointsSW_X[n+2] <- bignumber # Inf
    ROCpointsSW_Y[n+2] <- 0

    if (plot_curve) {
      if (DRAW_CIRCLES) {
        mytype <- "o"
      } else {
        mytype <- "l"
      }
      lines(ROCpointsSW_X/N, ROCpointsSW_Y/N, col=colour2, type=mytype, lty= my_lty)
    }

    AreaSW <- 0
    for(i in 3:(n+1)) {
      AreaSW <- AreaSW + ((-ROCpointsSW_Y[i]-ROCpointsSW_Y[i-1])/2)*(ROCpointsSW_X[i]-ROCpointsSW_X[i-1])
    }

#    print(ROCpointsSW_X)
#    print(ROCpointsSW_Y)

    print("Area (segmentwise, i.e., exact):")
    print(AreaSW)

    print("2*Area/n^2 (segmentwise, i.e., exact):")
    print(2*AreaSW/n^2)

    print("sample var*(n-1)/n  = population var:")
    print((var(errors)*(n-1)/n) )

    print("(var*(n-1)/n)  *n^2 / 2 = population var * n^2 / 2")
    print((var(errors)*(n-1)/n) *n ^2 / 2)
  }


  print("*************************************")

  if (SEGMENTWISEMETHOD)
    return(rbind(ROCpointsSW_X, ROCpointsSW_Y))  # Better, finer calculation if available
  else
    return(rbind(ROCpointsPW_X, ROCpointsPW_Y))  # Rough, approximate calculation if the other is not available
}











#### RROC_convexhull ####
#
# Returns convexhull of a set of points in RROC space and (optionally) plots it on a previous created RROC space
#
#### Arguments:
# x,y: the arrays for the (x,y) points in RROC space
# plot_hull: If TRUE, also plots the hull
# normalise_axes: divide x-axis and y-axis by N (only for plotting)
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
#
#### Returns
# The hull as a bidimensional matrix with the X,Y points of the curve.
#
RROC_convexhull <- function(x, y, plot_hull, normalise_axes=FALSE, N) {

  POINTS_ON_THE_HULL <- chull(x,y)       # Returns the indices of the points on the convex hull
  CONVEXHULLX <- x[POINTS_ON_THE_HULL]
  CONVEXHULLY <- y[POINTS_ON_THE_HULL]

  if (plot_hull) {
    if (!normalise_axes) {
      N <- 1 
    }
    lines(CONVEXHULLX/N, CONVEXHULLY/N, col="black", type='o', pch=4)
  }

  return(rbind(CONVEXHULLX, CONVEXHULLY))  
}








#### RROC_intercept ####
#
# Intercepts a cost context (alpha) on a RROC curve (assumes the curve is convex)
#
#### Arguments:
# CURVE: a RROC curve as an array of points (a matrix)
# alpha: the cost context (if alpha==0 it returns Best_intercept=-Inf, if alpha==1 it returns Best_intercept=0 by definition of the RROC space) 
# plot_isometric: TRUE if the isometric is plotted
# mycol: colour of the isometric (if plotted)
# normalise_axes: divide x-axis and y-axis by N (only for plotting)
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
#
#### Returns
# The intercept at x=0 and the index of the point that achieves it.
#
RROC_intercept <- function(CURVE, alpha, plot_isometric=TRUE, mycol="lightgrey", normalise_axes=FALSE, N=1)  {

  len <- length(CURVE[1,])

  Best_intercept <- -Inf
  if (alpha == 0) {
    slope <- 10000000 # This looks pretty vertical, eh?
    Best_i <- 0
  } else if (alpha == 1) {
    Best_intercept <- 0  
    Best_i <- len+1
  } else {
    slope <- (1-alpha)/alpha

    Best_i <- NULL
    for(i in 1: len) {
      x <- CURVE[1,i]
      y <- CURVE[2,i]
      intercept <- y - slope*x
      if (intercept > Best_intercept) {
        Best_intercept <- intercept
        Best_i <- i
      } else {
        break   # Since the curve is convex, we know we can stop now.
      }
    }
  }

  if (plot_isometric) {
    if (!normalise_axes) {
      N <- 1 
    }
    bi <- Best_intercept/N
    if (bi == -Inf) 
      bi <- -1000000000  # Instead of -Inf because it cannot be plotted
    abline(bi, slope, col=mycol)
  }

  list(Best_intercept, Best_i)
}





#### RROC_intercept_many ####
#
# Intercepts a cost context (alpha) on several RROC curve (as RROC_intercept_many) but for several curves in a list
#
#### Arguments:
# curve_list: a list of RROC curves (each of them as an array of points, a matrix)
# alpha: the cost context (if alpha==0 it returns Best_intercept=-Inf, if alpha==1 it returns Best_intercept=0 by definition of the RROC space) 
# plot_isometric: TRUE if the isometric is plotted
# mycol: colour of the isometric (if plotted)
# normalise_axes: divide x-axis and y-axis by N (only for plotting)
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
#
#### Returns
# The intercept at x=0 and the model that achieves it (if there are many returns the tie is solved by the order in the curve_list).
#
RROC_intercept_many <- function(curve_list, alpha, plot_isometric=TRUE, mycol="lightgrey", normalise_axes=FALSE, N=1)  {


  Best_intercept <- -Inf
  Best_model <- NULL
  if (alpha == 0) {
    slope <- 10000000 # This looks pretty vertical, eh?
  } else if (alpha == 1) {
    Best_intercept <- 0  
  } else {
    slope <- (1-alpha)/alpha

    num_curves <- length(curve_list)
    for(i in 1:num_curves) {
      res <- RROC_intercept(curve_list[[i]], alpha, plot_isometric=FALSE)
      intercept <- res[[1]]
      if (intercept > Best_intercept) {
        Best_intercept <- intercept
        Best_model <- i
      }
    }

    if (plot_isometric) {
      if (!normalise_axes) {
        N <- 1 
      }
      bi <- Best_intercept/N
      if (bi == -Inf) 
        bi <- -1000000000  # Instead of -Inf because it cannot be plotted
      abline(bi, slope, col=mycol)
    }
  }

  list(Best_intercept, Best_model)
}




#### RROC_dominance_intervals ####
#
# Finds the models that dominate (and their ranges) for several curves in a list
#
#### Arguments:
# curve_list: a list of RROC curves (each of them as an array of points, a matrix)
# RESOLUTION: number of alphas to be explored between 0 and 1. Each alpha is a cost context
#
#### Returns
# A list of intervals, each interval being (a,b,c), where [a,b] is the alpha interval and c is the index of the model.
#
RROC_dominance_intervals <- function(curve_list, RESOLUTION) {

  RESOLUTION <- 1000
  bestmodel_vect <- NULL
  dom_intervals <- NULL
  lastbm <- 0
  lastalpha <- 0
  alpha_vect <- seq(0,1, length.out = RESOLUTION)
  for(alpha in alpha_vect) { 
    reslist <- RROC_intercept_many(curve_list, alpha, plot_isometric=FALSE)
    bm <- reslist[[2]]
    if (is.null(bm))
      bm <- NA
    else if (bm != lastbm) {  # New interval of dominance
      if (lastbm !=0)  {
        dom_intervals <- c(dom_intervals, list(c(lastalpha,alpha,lastbm)))
        lastalpha <- alpha
      }
      lastbm <- bm
    }

    bestmodel_vect <- c(bestmodel_vect, bm)  # Store best model for this alpha
  }

  dom_intervals <- c(dom_intervals, list(c(lastalpha,1,lastbm)))

#  plot(alpha_vect,bestmodel_vect)
#  print(rbind(alpha_vect,bestmodel_vect))
#  print(dom_intervals)

  dom_intervals
}






#### loss_linlin ####
#
# Returns the asymmetric absolute error loss (linlin)
#
#### Arguments:
# predicted: an array of predicted values
# actual: an array of true values (size of predicted and actual should match)
# alpha: the cost context (between 0 and 1)
#
#### Returns
# The loss
#
loss_linlin <- function(predicted, actual, alpha) {

  errors <- predicted-actual                        # Calculates residuals (error vector)

  OVER <- sum(errors[which(errors > 0)])  # OVER
  UNDER <- sum(errors[which(errors <= 0)])  # UNDER

  if (alpha == 0) {
    cost <- 2*OVER
  } else if (alpha == 1) {
    cost <- -2*UNDER
  } else {
    cost <- -2*alpha*(UNDER) + 2*(1-alpha)*OVER  
  }
  cost
}



#### calibrate_regressor ####
#
# Calibrates regressor by adding a constant shift to get 0 bias (overpredictions = underpredictions)
#
#### Arguments:
# predicted: an array of predicted values
# actual: an array of true values (size of predicted and actual should match)
#
#### Returns
# The predictions shifted to have bias 0
#
calibrate_regressor <- function(predicted, actual) { # shifts model predicted to have 0 bias (overpredictions = underpredictions)
  orig_error <- predicted - actual
  predicted <- predicted - sum(orig_error)/length(predicted)
  predicted
}





#### optimal_shift_linlin ####
#
# Returns the best shift for a regression model (wrt. the linlin loss determined by an alpha)
# There are three ways of calculating the best shift:
#  index_opt: alpha and mycurve are not used (very efficient if we have previously calculated the intercept)
#  mycurve: alpha is needed. Uses a previously calculated curve (slightly more efficient than the following one)
#  only from alpha and predicted/actual: the usual way
#
#### Arguments:
# predicted: an array of predicted values
# actual: an array of true values (size of predicted and actual should match)
# alpha: the cost context (between 0 and 1)
# mycurve: the RROC curve values that may be used in the calculation
# index_opt: the index of the best intercept that may be used in the calculation
#
#### Returns
# The optimal shift
#
optimal_shift_linlin <- function(predicted, actual, alpha=0.5, mycurve=NULL, index_opt=NULL) {

  if (alpha == 0) {
    shift <- -Inf 
  } else if (alpha == 1) {
    shift <- Inf  
  } else {
    orig_error <- predicted-actual                        # Calculates residuals (error vector)
    errors <- sort(orig_error, decreasing=TRUE)  # sorts the errors

    if (!is.null(index_opt)) {
      shift <- -errors[index_opt]  
    } else {
      N <- length(orig_error)
      min_cost <- Inf
      for(i in 1:N) {
        s <- -errors[i]
        errors_shifted <- errors + s

        if (!is.null(mycurve)) {
          OVER <- mycurve[1,i+1]   # x -- OVER
          UNDER <- mycurve[2,i+1]   # y -- UNDER
        } else {
          OVER <- sum(errors_shifted[which(errors_shifted > 0)])  # OVER
          UNDER <- sum(errors_shifted[which(errors_shifted <= 0)])  # UNDER
        }

        cost <- -2*alpha*(UNDER) + 2*(1-alpha)*OVER  

        if (cost < min_cost) {
          shift <- s
          min_cost <- cost
        }
      }
    }
  }
  shift
}
  




#### optimal_shift_linlin_DEPRECATED ####
#
# An old way of calculating the best shift (inefficient)
# DEPRECATED: use optimal_shift_linlin insteas
#
optimal_shift_linlin_DEPRECATED <- function(predicted, actual, alpha, resolution) {
  ### This is still reasonably efficient if the number of examples is higher than the number of alphas

  if (alpha == 0) {
    shift <- -Inf 
  } else if (alpha == 1) {
    shift <- Inf  
  } else {

    orig_error <- predicted-actual                        # Calculates residuals (error vector)

    min_cost <- Inf
    for(s in seq(-max(orig_error),-min(orig_error),length.out=resolution)) {
      errors_shifted <- orig_error + s

      # The linlin loss function can be calculated in an aggregated way
      OVER <- sum(errors_shifted[which(errors_shifted > 0)])  # OVER
      UNDER <- sum(errors_shifted[which(errors_shifted <= 0)])  # UNDER
      cost <- 2*alpha*(-UNDER) + 2*(1-alpha)*OVER  
 
      if (cost < min_cost) {
        shift <- s
        min_cost <- cost
      }
    }
  }
  shift
}










#### RCOST_space ####
#
# Plots the RCOST space
#
#### Arguments:
# ylim0: part of the space shown
# normalise_axes: divide x-axis and y-axis by N
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
#
RCOST_space_linlin <- function(ylim0, normalise_axes= FALSE, N=NULL) {

  xlim0 <- 1
  xlab <- expression(alpha)  # alpha
  if (normalise_axes) {
    ylab <- "loss/N"
  } else {
    ylab <- "loss"
    N <- 1
  }

  plot(0, 0, lty="blank", col="white", xlim=c(0,xlim0), ylim=(c(0,ylim0/N)), xlab=xlab, ylab=ylab)  
}




#### RCOST_curve ####
#
# Returns a RCOST curve of a regression model and (optionally) plots it on a previous created RCOST space
#
#### Arguments:
# predicted: an array of predicted values
# actual: an array of true values (size of predicted and actual should match)
# alpha_vect: this is the resolution of the x-axis, i.e., the values between 0 and 1 that we want to plot.
# shift_vect: the previously derived shifts that must be used for each alpha.  (if NULL, the predicted vector is not shifted)
# plot_curve: if TRUE, plots the RCOST curve on a previously created RCOST curve. Set to FALSE if you only want to get the points of curve.
# colour2: colour used for the curve
# DRAW_CIRCLES: if TRUE it shows a small circle for each point. It is TRUE by default.
# my_lty : line style for the curve
# my_lwd : line width for the curve
# normalise_axes: divide x-axis and y-axis by N (only for plotting)
# N: Value used to normalise the axes (if (normalise_axes==FALSE) this value is ignored). Usually, for normalised curves, N has to be set to length(predicted).
#
#### Returns
# The curve as a loss array corresponding to the alpha_vect.
#

# Regression Cost Curves
RCOST_curve_linlin <- function(predicted, actual, alpha_vect, shift_vect=NULL, plot_curve, colour2, DRAW_CIRCLES= TRUE, my_lty=1, my_lwd=1, normalise_axes= FALSE, N=NULL)  {

  if (!normalise_axes) {
    N <- 1
  }


  nalphas <- length(alpha_vect)
  if (is.null(shift_vect)) {
    shift_vect <- rep(0,nalphas)
  }
 
  loss_vect <- NULL
  i <- 1
  for (alpha in alpha_vect) {
    res <- loss_linlin(predicted+shift_vect[i],actual,alpha)
    loss_vect <- c(loss_vect,res)
    i <- i+1
  }

#  print(rbind(alpha_vect, shift_vect, loss_vect))

  if (plot_curve) {
    if (DRAW_CIRCLES) {
      mytype <- "o"
    } else {
      mytype <- "l"
    }
    lines(alpha_vect, loss_vect/N, col=colour2, type=mytype, lty=my_lty, lwd=my_lwd)
  }

  loss_vect
}










#### abserror ####
#
# Calculates the absolute error between two values
#
#### Arguments:
# x, y: the values
#
#### Returns
# The absolute difference
#
abserror <- function(x,y) {
  abs(x - y)
}


#### squerror ####
#
# Calculates the squared error between two values
#
#### Arguments:
# x, y: the values
#
#### Returns
# The squared error
#
squerror <- function(x,y) {
  (x - y)^2
}





#### REC_space ####
#
# Plots the REC space
#
#### Arguments:
# xlim0: the width of the plot
# REC_ERROR_FUNCTION: error function used
#

REC_space <- function(xlim0, REC_ERROR_FUNCTION) {

  if (REC_ERROR_FUNCTION == 1) {
    xlab <- "Absolute deviation tolerance"
  } else {
    xlab <- "Squared error tolerance"
  }
  ylim0 <- 1
  ylab <- "Accuracy"

  plot(0, 0, lty="blank", col="white", xlim=c(0,xlim0), ylim=(c(0,ylim0)), xlab=xlab, ylab=ylab)  
}




#### REC_curve ####
#
# Returns a Regression Error Characteristic (REC) curve of a regression model and (optionally) plots it on a previous created REC space
#
#### Arguments:
# predicted: an array of predicted values
# actual: an array of true values (size of predicted and actual should match)
# errorfun: which error function is used (typically the absolute deviation (error) or the squared error are used
# plot_curve: if TRUE, plots the RROC curve on a previously created RROC curve. Set to FALSE if you only want to get the points of curve.
# colour2: colour used for the curve
# DRAW_CIRCLES: if TRUE it shows a small circle for each point. It is TRUE by default.
#
#### Returns
# The curve as a bidimensional matrix with the X,Y points of the curve.

REC_curve <- function(predicted, actual, errorfun, plot_curve, colour2, DRAW_CIRCLES= TRUE, my_lty=1)  {


  orig_error <- predicted-actual                        # Calculates residuals (error vector)

  print("*************************************")
  print("REC_curve function: Regression model:")
#  print(rbind(predicted,actual,orig_error))
  
  mae <- mean(abs(orig_error))                          # Calculates Mean Absolute Error
  mse <- mean(orig_error^2)                             # Calculates Mean Squared Error

  print("mae and mse")
  print(mae)
  print(mse)
  print("bias")
  print(mean(orig_error))


  errors <- errorfun(predicted,actual)              # Calculates error vector using errorfun

  errors <- sort(errors, decreasing=FALSE)  # sorts the errors
  n <- length(errors)
  eprev <- 0
  correct <- 0
  
  # Draws the curve from bottom-left corner to top-right corner
  RECpoints_X <- NULL
  RECpoints_Y <- NULL
  RECpoints_X[1] <- 0
  RECpoints_Y[1] <- 0
  
  for(i in 1:(n)) {
    if (errors[i] > eprev) {
      RECpoints_X[i+1] <- eprev
      RECpoints_Y[i+1] <- correct / n
      eprev <- errors[i]
    }
    correct <- correct + 1
  }
  RECpoints_X[i+1] <- errors[i]
  RECpoints_Y[i+1] <- correct / n

  RECpoints_X[i+2] <- errors[i]*10000 # To ensure that it reaches the end (I can't put infinite because it wouldn't be plotted)
  RECpoints_Y[i+2] <- 1

  if (plot_curve) {
    if (DRAW_CIRCLES) {
      mytype <- "o"
    } else {
      mytype <- "l"
    }
    lines(RECpoints_X, RECpoints_Y, col=colour2, type=mytype, lty=my_lty)
  }

  Area <- 0
  for(i in 2:(n+1)) {
    Area <- Area + ((1-(RECpoints_Y[i]+RECpoints_Y[i-1])/2))*(RECpoints_X[i]-RECpoints_X[i-1])
  }

  print("Area (segmentwise, i.e., exact):")
  print(Area)

  print("error var:")
  print((var(orig_error)*(n-1)/n) )

  print("(error_var*(n-1)/n)  *n^2 / 2:")
  print((var(orig_error)*(n-1)/n) *n ^2 / 2)


  print("*************************************")

  return(rbind(RECpoints_X, RECpoints_Y))  

}














